@@ -65,12 +65,12 @@ struct _TerminalScreenPrivate
   char *cooked_title, *cooked_icon_title;
   char *override_title;
   gboolean icon_title_set;
+  char *initial_working_directory;
   char **initial_env;
   char **override_command;
-  char *working_dir;
   int child_pid;
+  int pty_fd;
   double font_scale;
-  guint recheck_working_dir_idle;
   gboolean user_title; /* title was manually set */
   GSList *match_tags;
 };
@@ -104,8 +104,6 @@ enum
   TARGET_TAB
 };
 
-static void terminal_screen_init        (TerminalScreen      *screen);
-static void terminal_screen_class_init  (TerminalScreenClass *klass);
 static void terminal_screen_dispose     (GObject             *object);
 static void terminal_screen_finalize    (GObject             *object);
 static void terminal_screen_drag_data_received (GtkWidget        *widget,
@@ -136,8 +134,6 @@ static gboolean terminal_screen_format_title (TerminalScreen *screen, const char
 static void terminal_screen_cook_title      (TerminalScreen *screen);
 static void terminal_screen_cook_icon_title (TerminalScreen *screen);
 
-static void queue_recheck_working_dir (TerminalScreen *screen);
-
 static char* terminal_screen_check_match       (TerminalScreen            *screen,
                                                 int                   column,
                                                 int                   row,
@@ -153,9 +149,10 @@ static guint signals[LAST_SIGNAL];
 #define HOST HOSTCHARS_CLASS "+(\\." HOSTCHARS_CLASS "+)*"
 #define PORT "(?:\\:[[:digit:]]{1,5})?"
 #define PATHCHARS_CLASS "[-[:alnum:]\\Q_$.+!*,;@&=?/~#%\\E]"
+#define PATHTERM_CLASS "[^\\Q]'.}>) \t\r\n,\"\\E]"
 #define SCHEME "(?:news:|telnet:|nntp:|file:\\/|https?:|ftps?:|webcal:)"
 #define USERPASS USERCHARS_CLASS "+(?:" PASSCHARS_CLASS "+)?"
-#define URLPATH   "(/"PATHCHARS_CLASS"+(?:[(]"PATHCHARS_CLASS"*[)])*"PATHCHARS_CLASS"*)*"
+#define URLPATH   "(?:(/"PATHCHARS_CLASS"+(?:[(]"PATHCHARS_CLASS"*[)])*"PATHCHARS_CLASS"*)*"PATHTERM_CLASS")?"
 
 typedef struct {
   const char *pattern;
@@ -184,6 +181,59 @@ static guint n_skey_regexes;
 
 G_DEFINE_TYPE (TerminalScreen, terminal_screen, VTE_TYPE_TERMINAL)
 
+static char *
+cwd_of_pid (int pid)
+{
+  static const char patterns[][18] = {
+    "/proc/%d/cwd",         /* Linux */
+    "/proc/%d/path/cwd",    /* Solaris >= 10 */
+  };
+  guint i;
+  
+  if (pid == -1)
+    return NULL;
+
+  /* Try to get the working directory using various OS-specific mechanisms */
+  for (i = 0; i < G_N_ELEMENTS (patterns); ++i)
+    {
+      char cwd_file[64];
+      char buf[PATH_MAX + 1];
+      int len;
+
+      g_snprintf (cwd_file, sizeof (cwd_file), patterns[i], pid);
+      len = readlink (cwd_file, buf, sizeof (buf) - 1);
+
+      if (len > 0 && buf[0] == '/')
+        return g_strndup (buf, len);
+
+      /* If that didn't do it, try this hack */
+      if (len <= 0)
+        {
+          char *cwd, *working_dir = NULL;
+
+          cwd = g_get_current_dir ();
+          if (cwd != NULL)
+            {
+              /* On Solaris, readlink returns an empty string, but the
+               * link can be used as a directory, including as a target
+               * of chdir().
+               */
+              if (chdir (cwd_file) == 0)
+                {
+                  working_dir = g_get_current_dir ();
+                  chdir (cwd);
+                }
+              g_free (cwd);
+            }
+
+          if (working_dir)
+            return working_dir;
+        }
+    }
+
+  return NULL;
+}
+
 static void
 free_tag_data (TagData *tagdata)
 {
@@ -282,9 +332,6 @@ terminal_screen_realize (GtkWidget *widget)
   vte_terminal_set_background_transparent (VTE_TERMINAL (screen),
                                            bg_type == TERMINAL_BACKGROUND_TRANSPARENT &&
                                            !terminal_window_uses_argb_visual (priv->window));
-
-  /* FIXME: why do this on realize? */
-  terminal_window_set_size (priv->window, screen, TRUE);
 }
 
 static void
@@ -345,12 +392,8 @@ terminal_screen_init (TerminalScreen *screen)
 
   vte_terminal_set_mouse_autohide (VTE_TERMINAL (screen), TRUE);
 
-  priv->working_dir = g_get_current_dir ();
-  if (priv->working_dir == NULL) /* shouldn't ever happen */
-    priv->working_dir = g_strdup (g_get_home_dir ());
   priv->child_pid = -1;
-
-  priv->recheck_working_dir_idle = 0;
+  priv->pty_fd = -1;
 
   priv->font_scale = PANGO_SCALE_MEDIUM;
 
@@ -562,7 +605,7 @@ terminal_screen_class_init (TerminalScreenClass *klass)
 
   g_object_class_install_property
     (object_class,
-     PROP_ICON_TITLE,
+     PROP_OVERRIDE_COMMAND,
      g_param_spec_boxed ("override-command", NULL, NULL,
                          G_TYPE_STRV,
                          G_PARAM_READWRITE | G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK | G_PARAM_STATIC_BLURB));
@@ -651,16 +694,13 @@ terminal_screen_finalize (GObject *object)
 
   terminal_screen_set_profile (screen, NULL);
 
-  if (priv->recheck_working_dir_idle)
-    g_source_remove (priv->recheck_working_dir_idle);
-  
   g_free (priv->raw_title);
   g_free (priv->cooked_title);
   g_free (priv->override_title);
   g_free (priv->raw_icon_title);
   g_free (priv->cooked_icon_title);
+  g_free (priv->initial_working_directory);
   g_strfreev (priv->override_command);
-  g_free (priv->working_dir);
   g_strfreev (priv->initial_env);
 
   g_slist_foreach (priv->match_tags, (GFunc) free_tag_data, NULL);
@@ -678,18 +718,19 @@ terminal_screen_new (TerminalProfile *profile,
                      double           zoom)
 {
   TerminalScreen *screen;
+  TerminalScreenPrivate *priv;
 
   g_return_val_if_fail (TERMINAL_IS_PROFILE (profile), NULL);
 
   screen = g_object_new (TERMINAL_TYPE_SCREEN, NULL);
+  priv = screen->priv;
 
   terminal_screen_set_profile (screen, profile);
 
   if (title)
     terminal_screen_set_override_title (screen, title);
 
-  if (working_dir)
-    terminal_screen_set_working_dir (screen, working_dir);
+  priv->initial_working_directory = g_strdup (working_dir);
 
   if (override_command)
     terminal_screen_set_override_command (screen, override_command);
@@ -1073,9 +1114,14 @@ terminal_screen_set_font (TerminalScreen *screen)
     g_object_get (profile, TERMINAL_PROFILE_FONT, &desc, NULL);
   g_assert (desc);
 
-  pango_font_description_set_size (desc,
-				   priv->font_scale *
-				   pango_font_description_get_size (desc));
+  if (pango_font_description_get_size_is_absolute (desc))
+    pango_font_description_set_absolute_size (desc,
+                                              priv->font_scale *
+                                              pango_font_description_get_size (desc));
+  else
+    pango_font_description_set_size (desc,
+                                     priv->font_scale *
+                                     pango_font_description_get_size (desc));
 
   vte_terminal_set_font (VTE_TERMINAL (screen), desc);
 
@@ -1226,12 +1272,13 @@ show_command_error_dialog (TerminalScreen *screen,
 {
   g_assert (error != NULL);
   
-  terminal_util_show_error_dialog ((GtkWindow*) gtk_widget_get_ancestor (GTK_WIDGET (screen), GTK_TYPE_WINDOW), NULL,
-                                   _("There was a problem with the command for this terminal: %s"), error->message);
+  terminal_util_show_error_dialog ((GtkWindow*) gtk_widget_get_ancestor (GTK_WIDGET (screen), GTK_TYPE_WINDOW), NULL, error,
+                                   "%s", _("There was a problem with the command for this terminal"));
 }
 
 static gboolean
 get_child_command (TerminalScreen *screen,
+                   const char     *shell_env,
                    char          **file_p,
                    char         ***argv_p,
                    GError        **err)
@@ -1270,7 +1317,7 @@ get_child_command (TerminalScreen *screen,
       const char *only_name;
       char *shell;
 
-      shell = egg_shell ();
+      shell = egg_shell (shell_env);
 
       file = g_strdup (shell);
       
@@ -1306,7 +1353,8 @@ get_child_command (TerminalScreen *screen,
 }
 
 static char**
-get_child_environment (TerminalScreen *screen)
+get_child_environment (TerminalScreen *screen,
+                       char **shell)
 {
   TerminalScreenPrivate *priv = screen->priv;
   GtkWidget *term = GTK_WIDGET (screen);
@@ -1367,7 +1415,7 @@ get_child_environment (TerminalScreen *screen)
   g_free (proxymode);
 
   /* Do we already have a proxy setting? */
-  if (getenv ("http_proxy"))
+  if (g_hash_table_lookup (env_table, "http_proxy") != NULL)
     use_proxy = FALSE;
 
   /* Do we have no proxy host or an empty string? */
@@ -1477,6 +1525,8 @@ get_child_environment (TerminalScreen *screen)
     g_ptr_array_add (retval, g_strdup_printf ("%s=%s", e, v ? v : ""));
   g_ptr_array_add (retval, NULL);
 
+  *shell = g_strdup (g_hash_table_lookup (env_table, "SHELL"));
+
   g_hash_table_destroy (env_table);
   return (char **) g_ptr_array_free (retval, FALSE);
 }
@@ -1485,31 +1535,40 @@ void
 terminal_screen_launch_child (TerminalScreen *screen)
 {
   TerminalScreenPrivate *priv = screen->priv;
+  VteTerminal *terminal = VTE_TERMINAL (screen);
   TerminalProfile *profile;
   char **env, **argv;
-  char *path;
-  GError *err;
+  char *path, *shell = NULL;
+  GError *err = NULL;
   gboolean update_records;
+  const char *working_dir;
 
   profile = priv->profile;
 
-  err = NULL;
-  if (!get_child_command (screen, &path, &argv, &err))
+  env = get_child_environment (screen, &shell);
+
+  if (!get_child_command (screen, shell, &path, &argv, &err))
     {
       show_command_error_dialog (screen, err);
       g_error_free (err);
+
+      g_strfreev (env);
+      g_free (shell);
       return;
     }
 
-  env = get_child_environment (screen);
-
   update_records = terminal_profile_get_property_boolean (profile, TERMINAL_PROFILE_UPDATE_RECORDS);
 
-  priv->child_pid = vte_terminal_fork_command (VTE_TERMINAL (screen),
+  if (priv->initial_working_directory)
+    working_dir = priv->initial_working_directory;
+  else
+    working_dir = g_get_home_dir ();
+
+  priv->child_pid = vte_terminal_fork_command (terminal,
                                                path,
                                                argv,
                                                env,
-                                               terminal_screen_get_working_dir (screen),
+                                               working_dir,
                                                terminal_profile_get_property_boolean (profile, TERMINAL_PROFILE_LOGIN_SHELL),
                                                update_records,
                                                update_records);
@@ -1518,9 +1577,13 @@ terminal_screen_launch_child (TerminalScreen *screen)
     {
 
       terminal_util_show_error_dialog ((GtkWindow*) gtk_widget_get_ancestor (GTK_WIDGET (screen), GTK_TYPE_WINDOW), NULL,
+                                       err,
                                        "%s", _("There was an error creating the child process for this terminal"));
     }
   
+  priv->pty_fd = vte_terminal_get_pty (terminal);
+
+  g_free (shell);
   g_free (path);
   g_strfreev (argv);
   g_strfreev (env);
@@ -1586,19 +1649,19 @@ terminal_screen_button_press (GtkWidget      *widget,
   TerminalScreenPrivate *priv = screen->priv;
   gboolean (* button_press_event) (GtkWidget*, GdkEventButton*) =
     GTK_WIDGET_CLASS (terminal_screen_parent_class)->button_press_event;
-  int char_width, char_height, x, y, xpad, ypad;
+  int char_width, char_height, row, col, xpad_total, ypad_total;
   char *matched_string;
   int matched_flavor = 0;
   guint state;
 
   state = event->state & gtk_accelerator_get_default_mod_mask ();
 
-  vte_terminal_get_padding (VTE_TERMINAL (screen), &xpad, &ypad);
   terminal_screen_get_cell_size (screen, &char_width, &char_height);
+  vte_terminal_get_padding (VTE_TERMINAL (screen), &xpad_total, &ypad_total);
 
-  x = (event->x - xpad) / char_width;
-  y = (event->y - ypad) / char_height;
-  matched_string = terminal_screen_check_match (screen, x, y, &matched_flavor);
+  row = (event->x - xpad_total / 2) / char_width;
+  col = (event->y - ypad_total / 2) / char_height;
+  matched_string = terminal_screen_check_match (screen, row, col, &matched_flavor);
   
   if (matched_string != NULL &&
       (event->button == 1 || event->button == 2) &&
@@ -1726,112 +1789,60 @@ terminal_screen_get_dynamic_icon_title (TerminalScreen *screen)
   return screen->priv->raw_icon_title;
 }
 
-void
-terminal_screen_set_working_dir (TerminalScreen *screen,
-                                 const char     *dirname)
+/**
+ * terminal_screen_get_current_dir:
+ * @screen:
+ *
+ * Tries to determine the current working directory of the foreground process
+ * in @screen's PTY, falling back to the current working directory of the
+ * primary child.
+ * 
+ * Returns: a newly allocated string containing the current working directory,
+ *   or %NULL on failure
+ */
+char*
+terminal_screen_get_current_dir (TerminalScreen *screen)
 {
   TerminalScreenPrivate *priv = screen->priv;
-  
-  g_return_if_fail (TERMINAL_IS_SCREEN (screen));
+  char *cwd;
 
-  g_free (priv->working_dir);
-  priv->working_dir = g_strdup (dirname);
+  if (priv->pty_fd != -1) {
+#if 0
+    /* Get the foreground process ID */
+    cwd = cwd_of_pid (tcgetpgrp (priv->pty_fd));
+    if (cwd != NULL)
+      return cwd;
+#endif
+
+    /* If that didn't work, try falling back to the primary child. See bug #575184. */
+    cwd = cwd_of_pid (priv->child_pid);
+    if (cwd != NULL)
+      return cwd;
+  }
+
+  return NULL;
 }
 
-const char*
-terminal_screen_get_working_dir (TerminalScreen *screen)
+/**
+ * terminal_screen_get_current_dir_with_fallback:
+ * @screen:
+ *
+ * Like terminal_screen_get_current_dir(), but falls back to returning
+ * @screen's initial working directory, with a further fallback to the
+ * user's home directory.
+ * 
+ * Returns: a newly allocated string containing the current working directory,
+ *   or %NULL on failure
+ */
+char*
+terminal_screen_get_current_dir_with_fallback (TerminalScreen *screen)
 {
   TerminalScreenPrivate *priv = screen->priv;
-  
-  g_return_val_if_fail (TERMINAL_IS_SCREEN (screen), NULL);
 
-  /* Try to update the working dir using various OS-specific mechanisms */
-  if (priv->child_pid >= 0)
-    {
-      static const char *patterns[] = {
-        "/proc/%d/cwd",         /* Linux */
-        "/proc/%d/path/cwd",    /* Solaris >= 10 */
-      };
-      char *file = NULL;
-      char buf[PATH_MAX+1];
-      int len = 0;
-      guint i;
+  if (priv->pty_fd == -1)
+    return g_strdup (priv->initial_working_directory);
 
-      /* First try to update the working dir using various OS-specific mechanisms */
-      file = NULL;
-      for (i = 0; i < G_N_ELEMENTS (patterns); ++i)
-        {
-          g_free (file);
-          file = g_strdup_printf (patterns[i], priv->child_pid);
-          len = readlink (file, buf, sizeof (buf) - 1);
-
-          if (len > 0 && buf[0] == '/')
-            {
-              buf[len] = '\0';
-              
-              g_free (priv->working_dir);
-              priv->working_dir = g_strdup (buf);
-
-              break;
-            }
-        }
-
-      /* If that did not do it, be bold */
-      if (len <= 0)
-        {
-          char *cwd;
-
-          cwd = g_get_current_dir ();
-          if (cwd != NULL)
-            {
-              /* On Solaris, readlink returns an empty string, but the
-               * link can be used as a directory, including as a target
-               * of chdir().
-               */
-              if (chdir (file) == 0)
-                {
-                  g_free (priv->working_dir);
-                  priv->working_dir = g_get_current_dir ();
-                  chdir (cwd);
-                }
-              g_free (cwd);
-            }
-        }
-
-      g_free (file);
-    }
-
-  return priv->working_dir;
-}
-
-static gboolean
-recheck_dir (void *data)
-{
-  TerminalScreen *screen = data;
-  TerminalScreenPrivate *priv = screen->priv;
-
-  priv->recheck_working_dir_idle = 0;
-  
-  /* called just for side effect */
-  terminal_screen_get_working_dir (screen);
-
-  /* remove idle */
-  return FALSE;
-}
-
-static void
-queue_recheck_working_dir (TerminalScreen *screen)
-{
-  TerminalScreenPrivate *priv = screen->priv;
-  
-  if (priv->recheck_working_dir_idle == 0)
-    {
-      priv->recheck_working_dir_idle =
-        g_idle_add_full (G_PRIORITY_LOW + 50,
-                         recheck_dir,
-                         screen,
-                         NULL);
-    }
+  return terminal_screen_get_current_dir (screen);
 }
 
 void
@@ -1871,8 +1882,6 @@ terminal_screen_window_title_changed (VteTerminal *vte_terminal,
   terminal_screen_set_dynamic_title (screen,
                                      vte_terminal_get_window_title (vte_terminal),
 				     FALSE);
-
-  queue_recheck_working_dir (screen);
 }
 
 static void
@@ -1881,9 +1890,7 @@ terminal_screen_icon_title_changed (VteTerminal *vte_terminal,
 {
   terminal_screen_set_dynamic_icon_title (screen,
                                           vte_terminal_get_icon_title (vte_terminal),
-					  FALSE);  
-
-  queue_recheck_working_dir (screen);
+					  FALSE);
 }
 
 static void
@@ -1896,6 +1903,7 @@ terminal_screen_child_exited (VteTerminal *terminal)
   /* No need to chain up to VteTerminalClass::child_exited since it's NULL */
 
   priv->child_pid = -1;
+  priv->pty_fd = -1;
   
   action = terminal_profile_get_property_enum (priv->profile, TERMINAL_PROFILE_EXIT_ACTION);
   
@@ -1908,6 +1916,7 @@ terminal_screen_child_exited (VteTerminal *terminal)
       terminal_screen_launch_child (screen);
       break;
     case TERMINAL_EXIT_HOLD:
+    default:
       break;
     }
 }
@@ -1937,7 +1946,7 @@ terminal_screen_drag_data_received (GtkWidget        *widget,
                                     gint              y,
                                     GtkSelectionData *selection_data,
                                     guint             info,
-                                    guint             time)
+                                    guint             timestamp)
 {
   TerminalScreen *screen = TERMINAL_SCREEN (widget);
   TerminalScreenPrivate *priv = screen->priv;
@@ -2144,7 +2153,7 @@ terminal_screen_drag_data_received (GtkWidget        *widget,
                                           GTK_WIDGET (screen));
         terminal_window_move_screen (source_window, dest_window, moving_screen, page_num + 1);
 
-        gtk_drag_finish (context, TRUE, TRUE, time);
+        gtk_drag_finish (context, TRUE, TRUE, timestamp);
       }
       break;
 
@@ -2266,7 +2275,8 @@ terminal_screen_save_config (TerminalScreen *screen,
   TerminalScreenPrivate *priv = screen->priv;
   VteTerminal *terminal = VTE_TERMINAL (screen);
   TerminalProfile *profile = priv->profile;
-  const char *profile_id, *dir;
+  const char *profile_id;
+  char *working_directory;
 
   profile_id = terminal_profile_get_property_string (profile, TERMINAL_PROFILE_NAME);
   g_key_file_set_string (key_file, group, TERMINAL_CONFIG_TERMINAL_PROP_PROFILE_ID, profile_id);
@@ -2278,9 +2288,11 @@ terminal_screen_save_config (TerminalScreen *screen,
   if (priv->override_title)
     g_key_file_set_string (key_file, group, TERMINAL_CONFIG_TERMINAL_PROP_TITLE, priv->override_title);
 
-  dir = terminal_screen_get_working_dir (screen);
-  if (dir != NULL && *dir != '\0') /* should always be TRUE anyhow */
-    terminal_util_key_file_set_string_escape (key_file, group, TERMINAL_CONFIG_TERMINAL_PROP_WORKING_DIRECTORY, dir);
+  /* FIXMEchpe: use the initial_working_directory instead?? */
+  working_directory = terminal_screen_get_current_dir (screen);
+  if (working_directory)
+    terminal_util_key_file_set_string_escape (key_file, group, TERMINAL_CONFIG_TERMINAL_PROP_WORKING_DIRECTORY, working_directory);
+  g_free (working_directory);
 
   g_key_file_set_double (key_file, group, TERMINAL_CONFIG_TERMINAL_PROP_ZOOM, priv->font_scale);
 
@@ -2303,13 +2315,12 @@ gboolean
 terminal_screen_has_foreground_process (TerminalScreen *screen)
 {
   TerminalScreenPrivate *priv = screen->priv;
-  int master_fd, fgpid;
+  int fgpid;
 
-  master_fd = vte_terminal_get_pty (VTE_TERMINAL (screen));
-  if (master_fd == -1)
+  if (priv->pty_fd == -1)
     return FALSE;
 
-  fgpid = tcgetpgrp (master_fd);
+  fgpid = tcgetpgrp (priv->pty_fd);
   if (fgpid == -1 || fgpid == priv->child_pid)
     return FALSE;
 
